# Copyright 2023-2024 Dominik Sekotill <dom.sekotill@kodo.org.uk>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Filter decorators for marking the requested protocol options and actions used
"""

from __future__ import annotations

from collections import defaultdict
from enum import Flag
from enum import auto
from typing import Callable
from typing import Literal
from typing import NamedTuple

from kilter.protocol.messages import ActionFlags
from kilter.protocol.messages import ProtocolFlags
from kilter.protocol.messages import Stage

from .session import Filter

__all__ = [
	"CanRespond", "NEVER", "BEFORE", "DURING", "AFTER",
	"responds_to_connect", "examine_helo",
	"examine_sender", "examine_recipients",
	"examine_headers", "examine_body",
	"get_flags", "modify_flags",
	"get_macros", "request_macros",
]

Decorator = Callable[[Filter], Filter]
SIZES = Literal[ProtocolFlags.NONE, ProtocolFlags.MDS_256K, ProtocolFlags.MDS_1M]

FLAGS_ATTRIBUTE = "filter_flags"
MACRO_ATTRIBUTE = "filter_macros"

DEFAULT_UNSET = \
	ProtocolFlags.NO_CONNECT | ProtocolFlags.NO_HELO | \
	ProtocolFlags.NO_SENDER | ProtocolFlags.NO_RECIPIENT | \
	ProtocolFlags.NO_DATA | ProtocolFlags.NO_BODY | \
	ProtocolFlags.NO_HEADERS | ProtocolFlags.NO_END_OF_HEADERS | \
	ProtocolFlags.NO_UNKNOWN | \
	ProtocolFlags.NR_CONNECT | ProtocolFlags.NR_HELO | \
	ProtocolFlags.NR_SENDER | ProtocolFlags.NR_RECIPIENT | \
	ProtocolFlags.NR_DATA | ProtocolFlags.NR_BODY | \
	ProtocolFlags.NR_HEADER | ProtocolFlags.NR_END_OF_HEADERS | \
	ProtocolFlags.NR_UNKNOWN


class CanRespond(Flag):
	"""
	Flags for fine indication of which stages during message sending a filter may respond at

	Used with `examine_headers()` and `examine_body()` to further refine which stages during
	message header and body transfer will synchronously block until a response of some kind
	is received.
	"""

	NEVER = 0
	BEFORE = auto()
	DURING = auto()
	AFTER = auto()
	ALL = BEFORE|DURING|AFTER


NEVER = CanRespond.NEVER
BEFORE = CanRespond.BEFORE
DURING = CanRespond.DURING
AFTER = CanRespond.AFTER


class FlagsTuple(NamedTuple):

	unset_options: ProtocolFlags = ProtocolFlags.NONE
	set_options: ProtocolFlags = ProtocolFlags.NONE
	set_actions: ActionFlags = ActionFlags.NONE


def modify_flags(
	set_options: ProtocolFlags = ProtocolFlags.NONE,
	unset_options: ProtocolFlags = ProtocolFlags.NONE,
	set_actions: ActionFlags = ActionFlags.NONE,
) -> Decorator:
	"""
	Return a decorator that modifies the given flags on a decorated filter
	"""
	def decorator(filtr: Filter) -> Filter:
		_set_flags(filtr, set_options, unset_options, set_actions)
		return filtr
	return decorator


def get_flags(filtr: Filter) -> FlagsTuple:
	"""
	Return the flags attached to a filter
	"""
	default = FlagsTuple(unset_options=DEFAULT_UNSET, set_actions=ActionFlags.ALL)
	return _get_flags(filtr, default)


def _set_flags(
	filtr: Filter,
	set_options: ProtocolFlags = ProtocolFlags.NONE,
	unset_options: ProtocolFlags = ProtocolFlags.NONE,
	set_actions: ActionFlags = ActionFlags.NONE,
) -> None:
	flags = _get_flags(filtr, FlagsTuple())
	flags = FlagsTuple(
		flags.unset_options|unset_options,
		flags.set_options|set_options,
		flags.set_actions|set_actions,
	)
	setattr(filtr, FLAGS_ATTRIBUTE, flags)


def _get_flags(filtr: Filter, default: FlagsTuple) -> FlagsTuple:
	assert isinstance(getattr(filtr, FLAGS_ATTRIBUTE, default), FlagsTuple)
	return getattr(filtr, FLAGS_ATTRIBUTE, default)


def request_macros(stage: Stage, *names: str) -> Decorator:
	"""
	Return a decorator that adds the given macro requests to a decorated filter
	"""
	def decorator(filtr: Filter) -> Filter:
		_set_flags(filtr, set_actions=ActionFlags.SETSYMLIST)
		macros = get_macros(filtr)
		macros[stage].update(names)
		return filtr
	return decorator


def get_macros(filtr: Filter) -> defaultdict[Stage, set[str]]:
	"""
	Return the requested macros attached to a filter
	"""
	try:
		macros = getattr(filtr, MACRO_ATTRIBUTE)
	except AttributeError:
		setattr(filtr, MACRO_ATTRIBUTE, (macros := defaultdict(set)))
	assert isinstance(macros, defaultdict)
	return macros


def responds_to_connect() -> Decorator:
	"""
	Mark a filter as possibly delivering a non-continue response to Connect events
	"""
	return modify_flags(unset_options=ProtocolFlags.NR_CONNECT)


def examine_helo(
	can_respond: bool = False,
) -> Decorator:
	"""
	Mark a filter as needing to examine the HELO command

	If `can_respond` is `False` the filter runner will attempt to negotiate faster event
	delivery by disabling the need to respond to this event.
	"""
	unset = ProtocolFlags.NO_HELO
	if can_respond:
		unset |= ProtocolFlags.NR_HELO
	return modify_flags(unset_options=unset)


def examine_sender(
	can_respond: bool = False,
	can_replace: bool = False,
) -> Decorator:
	"""
	Mark a filter as needing to examine and optionally replace the RCPT FROM sender

	If `can_respond` is `False` the filter runner will attempt to negotiate faster event
	delivery by disabling the need to respond to this event.

	If `can_replace` is `True` but is not offered by the MTA an exception will be raised
	during negotiation and the filter will be disabled.
	"""
	unset = ProtocolFlags.NO_SENDER
	if can_respond:
		unset |= ProtocolFlags.NR_SENDER
	return modify_flags(
		unset_options=unset,
		set_actions=ActionFlags.CHANGE_FROM if can_replace else ActionFlags.NONE,
	)


def examine_recipients(
	can_respond: bool = False,
	can_add: bool = False,
	can_remove: bool = False,
	include_rejected: bool = False,
	with_parameters: bool = False,
) -> Decorator:
	"""
	Mark a filter as needing to examine and optionally modify the RCPT TO recipients

	If `can_respond` is `False` the filter runner will attempt to negotiate faster event
	delivery by disabling the need to respond to this event.

	If `include_rejected` is `True` the recipients available to the filter will include any
	that the MTA or another filter has already rejected.

	The option `with_parameters` enables the use of RFC-1425 [section 6] extensions for
	"MAIL" commands (ratified by RFC-5321) when adding recipients.  The specific details of
	any extension parameters will be dependent on the MTA.

	If a requested option or update action is not offered by the MTA an exception will be
	raised during negotiation and the filter will be disabled.
	"""
	unset = ProtocolFlags.NO_RECIPIENT
	opts = ProtocolFlags.NONE
	acts = ActionFlags.NONE
	if can_respond:
		unset |= ProtocolFlags.NR_RECIPIENT
	if can_add:
		acts |= ActionFlags.ADD_RECIPIENT
	if can_add and with_parameters:
		acts |= ActionFlags.ADD_RECIPIENT_PAR
	if can_remove:
		acts |= ActionFlags.DELETE_RECIPIENT
	if include_rejected:
		opts |= ProtocolFlags.REJECTED_RECIPIENT
	return modify_flags(unset_options=unset, set_options=opts, set_actions=acts)


def examine_headers(
	can_respond: bool|CanRespond = False,
	can_add: bool = False,
	can_modify: bool = False,
	leading_space: bool = False,
) -> Decorator:
	"""
	Mark a filter as needing to examine and optionally add or modify message headers

	If `can_respond` is `False` the filter runner will attempt to negotiate faster event
	delivery by disabling the need to respond to this event.

	If `leading_space` is `True` the headers will be delivered without any whitespace
	removed from values (i.e. after the separating colon). This is for filters which need
	the exact bytes contained in message headers.

	If a requested option or update action is not offered by the MTA an exception will be
	raised during negotiation and the filter will be disabled.
	"""
	unset = ProtocolFlags.NO_HEADERS
	opts = ProtocolFlags.NONE
	acts = ActionFlags.NONE
	if isinstance(can_respond, bool):
		can_respond = CanRespond.ALL if can_respond else CanRespond.NEVER
	if CanRespond.BEFORE in can_respond:
		unset |= ProtocolFlags.NO_DATA | ProtocolFlags.NR_DATA
	if CanRespond.DURING in can_respond:
		unset |= ProtocolFlags.NR_HEADER
	if CanRespond.AFTER in can_respond:
		unset |= ProtocolFlags.NO_END_OF_HEADERS | ProtocolFlags.NR_END_OF_HEADERS
	if can_add:
		acts |= ActionFlags.ADD_HEADERS
	if can_modify:
		acts |= ActionFlags.CHANGE_HEADERS
	if leading_space:
		opts |= ProtocolFlags.HEADER_LEADING_SPACE
	return modify_flags(unset_options=unset, set_options=opts, set_actions=acts)


def examine_body(
	can_respond: bool|CanRespond = False,
	can_replace: bool = False,
	data_size: SIZES = ProtocolFlags.NONE,
) -> Decorator:
	"""
	Mark a filter as needing to examine and optionally replace message bodies

	If `can_respond` is `False` the filter runner will attempt to negotiate faster event
	delivery by disabling the need to respond to this event.

	The `data_size` option is a hint, and does not guarantee that the message will be
	delivered in blocks of that size. If `ProtocolFlags.NONE` (the default) the MTA's
	default will be used.

	If `can_replace` is `True` but is not offered by the MTA an exception will be raised
	during negotiation and the filter will be disabled.
	"""
	unset = ProtocolFlags.NO_BODY
	if isinstance(can_respond, bool):
		can_respond = CanRespond.ALL if can_respond else CanRespond.NEVER
	if CanRespond.BEFORE in can_respond:
		unset |= ProtocolFlags.NO_END_OF_HEADERS | ProtocolFlags.NR_END_OF_HEADERS
	if CanRespond.DURING in can_respond:
		unset |= ProtocolFlags.NR_BODY
	# CanRespond.AFTER is implicit
	return modify_flags(
		unset_options=unset, set_options=data_size,
		set_actions=ActionFlags.CHANGE_BODY if can_replace else ActionFlags.NONE,
	)



# Copyright 2022 Dominik Sekotill <dom.sekotill@kodo.org.uk>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Common helper utilities
"""

# mypy: disallow-any-explicit=False

from __future__ import annotations

from collections.abc import AsyncIterator
from collections.abc import Callable
from contextlib import asynccontextmanager
from typing import Generic
from typing import Optional
from typing import TypeVar

import anyio

T = TypeVar("T")
Fn = TypeVar("Fn", bound=Callable[..., object])


class Broadcast(anyio.Condition, Generic[T]):
	"""
	A reliable, blocking message queue for delivering to multiple listening tasks

	Listeners must acquire the lock (by using the `Broadcast` instance as a context manager)
	before calling `Broadcast.receive()` or it will fail.  If a listener is repeatedly
	awaiting messages in a loop, the loop should be inside the locked context or messages
	may be lost to race conditions.
	"""

	def __init__(self) -> None:
		super().__init__()
		self.obj: Optional[T] = None
		self.exc: Optional[BaseException|type[BaseException]] = None

	async def pre_receive_hook(self) -> None:
		"""
		A hook for subclasses to inject synchronisation instructions before awaiting objects
		"""  # noqa: D401

	async def post_send_hook(self) -> None:
		"""
		A hook for subclasses to inject synchronisation instructions after sending objects
		"""  # noqa: D401

	async def shutdown_hook(self) -> None:
		"""
		A hook for subclasses to inject cleanup or synchronisation instructions on close

		Users must ensure this method is called, especially if using a subclass which
		implements it.
		"""  # noqa: D401

	async def abort(self, exc: BaseException|type[BaseException]) -> None:
		"""
		Send a notification to all listeners to abort by raising an exception
		"""
		async with self._ready():
			assert self.exc is None and self.obj is None
			self.exc = exc
			self.notify_all()
		await self._post()

	async def send(self, obj: T) -> None:
		"""
		Send a message object and block until all listeners have received it
		"""
		async with self._ready():
			assert self.exc is None and self.obj is None
			self.obj = obj
			self.notify_all()
		await self._post()

	@asynccontextmanager
	async def _ready(self) -> AsyncIterator[None]:
		while 1:
			await anyio.sleep(0.0)
			async with self:
				if self.obj is not None or self.exc is not None:
					continue
				yield
				return

	async def _post(self) -> None:
		await anyio.sleep(0.0)  # ensure listeners have opportunity to wait for locks
		await self.post_send_hook()

		# Ensure all listeners have had a chance to lock and process self.obj
		while 1:
			async with self:
				if self.statistics().lock_statistics.tasks_waiting:  # pragma: no-branch
					continue
				self.obj = self.exc = None
				break

	async def receive(self) -> T:
		"""
		Listen for a single message and return it once it arrives
		"""
		await self.pre_receive_hook()
		await self.wait()
		if self.exc is not None:
			raise self.exc
		assert self.obj is not None
		return self.obj


def qualname(func: Fn) -> str:
	"""
	Return a qualified name for a callable
	"""
	if func.__module__ == "__main__":
		return func.__qualname__
	return f"{func.__module__}.{func.__qualname__}"



# Copyright 2022-2023, 2025 Dominik Sekotill <dom.sekotill@kodo.org.uk>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Coordinate receiving and sending raw messages with a filter and Session object

The primary class in this module (`Runner`) is intended to be used with an
`anyio.abc.Listener`, which can be obtained, for instance, from
`anyio.create_tcp_listener()`.
"""

from __future__ import annotations

import enum
import logging
from collections import defaultdict
from collections.abc import Iterable
from typing import Final
from typing import TypeAlias
from warnings import warn

import anyio.abc
from async_generator import aclosing

from kilter.protocol.buffer import SimpleBuffer
from kilter.protocol.core import EventMessage
from kilter.protocol.core import FilterMessage
from kilter.protocol.core import FilterProtocol
from kilter.protocol.core import ResponseMessage
from kilter.protocol.messages import *

from .options import get_flags
from .options import get_macros
from .session import Aborted
from .session import Filter
from .session import FilterResponse
from .session import Session
from .util import Broadcast
from .util import qualname

__all__ = [
	"Runner",
	"NegotiationError",
]

FinalResponse: TypeAlias = FilterResponse | TemporaryFailure

kiB: Final = 2**10
MiB: Final = 2**20

_logger = logging.getLogger(__package__)


class NegotiationError(Exception):
	"""
	An error raised when MTAs are not compatible with the filter
	"""


class State(enum.Enum):

	CONNECTED = enum.auto()
	SESSION = enum.auto()
	SESSION_ABORTED = enum.auto()
	MESSAGE = enum.auto()
	MESSAGE_ABORTED = enum.auto()


class _Broadcast(Broadcast[EventMessage]):

	def __init__(self) -> None:
		super().__init__()
		self.task_status = list[anyio.abc.TaskStatus[None]]()

	async def shutdown_hook(self) -> None:
		await self.pre_receive_hook()

	async def pre_receive_hook(self) -> None:
		while self.task_status:
			self.task_status.pop().started()


class Sender:
	"""
	Concrete implementation of `kilter.service.session.Sender`
	"""

	def __init__(self, client: anyio.abc.ByteSendStream, proto: FilterProtocol) -> None:
		self.client = client
		self.proto = proto

	async def send(self, message: FilterMessage) -> None:
		"""
		Encode and send a message to the client stream
		"""
		buffer = SimpleBuffer(1*kiB)
		self.proto.write_to(buffer, message)
		await self.client.send(buffer[:])
		if __debug__:
			_logger.debug(f"sent: {message}")


class Runner:
	"""
	A filter runner that coordinates passing data between a stream and multiple filters

	Instances can be used as handlers that can be passed to `anyio.abc.Listener.serve()` or
	used with any `anyio.abc.ByteStream`.
	"""

	def __init__(self, *filters: Filter):
		if len(filters) == 0:  # pragma: no-cover
			raise TypeError("Runner requires at least one filter to run")
		self.filters = set(filters)
		if len(filters) != len(self.filters):
			warn("Repeated filters will only be run once", stacklevel=2)
		self.use_skip = True

	async def __call__(self, client: anyio.abc.ByteStream) -> None:
		"""
		Return an awaitable that starts and coordinates filters
		"""
		buff = SimpleBuffer(1*MiB)
		proto = FilterProtocol(abort_on_unknown=True)
		sender = Sender(client, proto)
		session = Session(sender, _Broadcast())
		runner = SessionRunner(session)
		state = State.CONNECTED

		async with (
			aclosing(client),
			anyio.create_task_group() as tasks,
		):
			while 1:
				try:
					buff[:] = await client.receive(buff.available)
				except (
					anyio.EndOfStream,
					anyio.ClosedResourceError,
					anyio.BrokenResourceError,
				):
					return
				for message in proto.read_from(buff):
					if __debug__:
						_logger.debug(f"received: {message}")

					# If previous message was Abort, restart filters for any non-Abort/Close
					# message
					if state in (State.SESSION_ABORTED, State.MESSAGE_ABORTED):
						if not isinstance(message, Abort|Close):
							await runner.start(self.filters, tasks)
						state = (
							State.CONNECTED if state == State.SESSION_ABORTED else
							State.SESSION
						)

					match message:
						case Negotiate():
							await sender.send(await self._negotiate(message))
							continue
						case Connect():
							_logger.info(f"Client connected from {message.hostname}")
							await session.deliver(message)
							await runner.start(self.filters, tasks)
							if proto.needs_response(message):
								await sender.send(await runner.check_response() or Continue())
							continue
						case Helo():
							state = State.SESSION
						case EnvelopeFrom():
							state = State.MESSAGE
						case Abort() if state in (State.SESSION, State.MESSAGE):
							state = (
								State.SESSION_ABORTED if state == State.SESSION else
								State.MESSAGE_ABORTED
							)
						case Abort():
							_logger.warning("Unexpected Abort received")
							state = State.CONNECTED
						case Close():
							tasks.cancel_scope.cancel()
							return

					skip_or_cont = await session.deliver(message)
					if not proto.needs_response(message):
						continue
					if (resp := await runner.check_response()):
						await sender.send(resp)
					elif self.use_skip:
						await sender.send(skip_or_cont())
					else:
						await sender.send(Continue())

	async def _negotiate(self, message: Negotiate) -> Negotiate:
		_logger.info("Negotiating with MTA")

		optmask = ProtocolFlags.NONE
		options = \
			ProtocolFlags.SKIP | \
			ProtocolFlags.NO_HELO | \
			ProtocolFlags.NO_SENDER | ProtocolFlags.NO_RECIPIENT | \
			ProtocolFlags.NO_DATA | ProtocolFlags.NO_BODY | \
			ProtocolFlags.NO_HEADERS | ProtocolFlags.NO_END_OF_HEADERS | \
			ProtocolFlags.NR_CONNECT | ProtocolFlags.NR_HELO | \
			ProtocolFlags.NR_SENDER | ProtocolFlags.NR_RECIPIENT | \
			ProtocolFlags.NR_DATA | ProtocolFlags.NR_BODY | \
			ProtocolFlags.NR_HEADER | ProtocolFlags.NR_END_OF_HEADERS
		actions = ActionFlags.NONE
		macros = defaultdict(set)

		options &= message.protocol_flags  # Remove unoffered initial flags, they are not required

		for filtr in self.filters:
			flags = get_flags(filtr)
			optmask |= flags.unset_options
			options |= flags.set_options
			actions |= flags.set_actions

			for stage, names in get_macros(filtr).items():
				macros[stage].update(names)

		options &= ~optmask

		if (missing_actions := actions & ~message.action_flags):
			raise NegotiationError(f"MTA does not accept {missing_actions}")

		if (missing_options := options & ~message.protocol_flags):
			raise NegotiationError(f"MTA does not offer {missing_options}")

		self.use_skip = ProtocolFlags.SKIP in options

		return Negotiate(6, actions, options, dict(macros))


class SessionRunner:

	def __init__(self, session: Session):
		self.session = session
		self.filters = dict[Filter, FinalResponse|None]()

	async def start(self, filters: Iterable[Filter], task_group: anyio.abc.TaskGroup) -> None:
		"""
		Run all the given filters in a task group

		The session MUST have been primed by the delivery of a Connect message beforehand or
		filters will be unable to access the connection details.
		"""
		_logger.debug("Starting filters")
		for flter in filters:
			await task_group.start(self.run_filter, flter)

	async def run_filter(
		self,
		flter: Filter,
		task_status: anyio.abc.TaskStatus[None],
	) -> None:
		"""
		Run a filter as a subtask in a task group

		A `Future` for returning the filter's response is added to the
		`SessionRunner.filter` dict.
		"""
		if flter in self.filters:
			raise RuntimeError
		self.filters[flter] = None

		async with self.session:
			assert isinstance(self.session.broadcast, _Broadcast)
			status_notifiers = self.session.broadcast.task_status
			status_notifiers.append(task_status)

			try:
				resp: FinalResponse = await flter(self.session)
			except Aborted:
				_logger.debug(f"Aborted filter {qualname(flter)}")
				del self.filters[flter]
				return
			except Exception:
				_logger.exception(f"Error in filter {qualname(flter)}")
				resp = TemporaryFailure()
			if not isinstance(resp, FinalResponse):
				warn(f"expected a valid response from {qualname(flter)}, got {resp}")  # type: ignore # Don't fully trust users…
				resp = TemporaryFailure()
			self.filters[flter] = resp
			if task_status in status_notifiers:
				status_notifiers.remove(task_status)
				task_status.started()

	async def check_response(self) -> ResponseMessage|None:
		assert self.filters, "no filters when checking for a response"
		response: ResponseMessage|None = None
		complete = list[Filter]()
		for flter, result in self.filters.items():
			# If a filter has not finished or no response is expected, continue without
			# removing from filter container; remove failed filters and filters that have
			# accepted; return a response for rejections;
			match result:
				case None:
					continue
				case Accept():
					_logger.info("Accept from %s, waiting for remaining", qualname(flter))
				case TemporaryFailure() as response:
					_logger.warning("Filter failed: %s", flter)
				case Reject()|Discard()|ReplyCode() as response:
					_logger.info("Returning response %s from %s", type(response).__name__, qualname(flter))
					complete[:] = self.filters
					break
				case msg:
					raise AssertionError(f"unexpected filter result: {msg}")
			complete.append(flter)
		for flter in complete:
			del self.filters[flter]
		return response if response else None if self.filters else Accept()



"""
High level, asynchronous framework for writing mail filters

Kilter is a framework for writing mail filters (known as "milters")
compatible with Sendmail and Postfix MTAs.  Unlike many previous milter implementations in
Python it is not simply bindings to the libmilter library (originally from the Sendmail
project).  The framework aims to provide Pythonic interfaces for implementing filters,
including leveraging coroutines instead of libmilter's callback-style interface.
"""

from ..protocol import ResponseMessage
from .runner import Runner
from .session import END
from .session import START
from .session import After
from .session import Before
from .session import Session

__version__ = "0.3.1"

__all__ = [
	"After",
	"Before",
	"END",
	"ResponseMessage",
	"Runner",
	"START",
	"Session",
]



# Copyright 2022-2025 Dominik Sekotill <dom.sekotill@kodo.org.uk>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Sessions are the kernel of a filter, providing it with an async API to access messages
"""

from __future__ import annotations

from collections.abc import AsyncGenerator
from collections.abc import AsyncIterator
from collections.abc import Sequence
from contextvars import ContextVar
from dataclasses import dataclass
from enum import Enum
from ipaddress import IPv4Address
from ipaddress import IPv6Address
from pathlib import Path
from types import TracebackType
from typing import AsyncContextManager
from typing import Literal
from typing import Protocol
from typing import TypeAlias
from warnings import warn

from typing_extensions import Self

from ..protocol.core import EditMessage
from ..protocol.core import EventMessage
from ..protocol.messages import *
from . import util

FilterResponse: TypeAlias = Accept | Reject | Discard | ReplyCode


class Aborted(BaseException):
	"""
	An exception for aborting filters on receipt of an Abort message
	"""


class Filter(Protocol):
	"""
	Filters are callables that accept a `Session` and return a response
	"""

	async def __call__(self, session: Session, /) -> FilterResponse: ...  # noqa: D102


class Sender(Protocol):
	"""
	Senders asynchronously handle sending messages with their "send" method
	"""

	async def send(self, message: EditMessage) -> None: ...  # noqa: D102


class Phase(int, Enum):
	"""
	Session phases indicate what messages to expect and are impacted by received messages

	Users should not generally need to use these values, however an understanding of the
	state-flow they represent is useful for understanding some error exception
	raised by `Session` methods.
	"""

	INIT = 0
	"""
	This phase is the pre-connected phase of a session; this phase will be completed before
	users see the session object.
	"""

	CONNECT = 1
	"""
	This phase is the starting phase of a session, during which a HELO/EHLO message may be
	awaited with `Session.helo()`.
	"""

	MAIL = 2
	"""
	This phase is entered after HELO/EHLO, during which a MAIL message may be awaited with
	`Session.envelope_from()`.  The `Session.extension()` method may also be used to get
	the raw MAIL command with any extension arguments, or any other extension commands
	that the MTA does not support (if the MTA supports passing these commands to
	a filter).
	"""

	ENVELOPE = 3
	"""
	This phase is entered after MAIL, during which any RCPT commands may be awaited with
	`Session.envelope_recipients()`.  The `Session.extension()` method may also be used to
	get the raw RCPT command with any extension arguments, or any other extension commands
	that the MTA does not support (if the MTA supports passing these commands to
	a filter).
	"""

	HEADERS = 4
	"""
	This phase is entered after a DATA command, while message headers are processed.
	Headers may be iterated as they arrive, or be collected for later through the
	`Session.headers` object.
	"""

	BODY = 5
	"""
	This phase is entered after a message's headers have been processed.  The raw message
	body may be iterated over in chunks through the `Session.body` object.
	"""

	POST = 6
	"""
	This phase is entered once a message's body has been completed (or skipped).  During
	this phase the message editing methods of a `Session` object or the `Session.headers`
	and `Session.body` objects may be used.
	"""


@dataclass
class Position:
	"""
	A base class for `Before` and `After`, this class is not intended to be used directly
	"""

	subject: Header|Literal["start"]|Literal["end"]


@dataclass
class Before(Position):
	"""
	Indicates a relative position preceding a subject `Header` in a header list

	See `HeadersAccessor.insert`.
	"""

	subject: Header


@dataclass
class After(Position):
	"""
	Indicates a relative position following a subject `Header` in a header list

	See `HeadersAccessor.insert`.
	"""

	subject: Header


START = Position("start")
"""
Indicates the start of a header list, before the first (current) header
"""

END = Position("end")
"""
Indicates the end of a header list, after the last (current) header
"""


class Session:
	"""
	The kernel of a filter, providing an API for filters to access messages from an MTA
	"""

	host: str
	"""
	A hostname from a reverse address lookup performed when a client connects

	If no name is found this value defaults to the standard presentation format for
	`Session.address` surrounded by "[" and "]", e.g. "[192.0.2.100]"
	"""

	address: IPv4Address|IPv6Address|Path|None
	"""
	The address of the connected client, or None if unknown
	"""

	port: int
	"""
	The port of the connected client if applicable, or 0 otherwise
	"""

	macros: dict[str, str]
	"""
	A mapping of string replacements sent by the MTA

	See `smfi_getsymval <https://pythonhosted.org/pymilter/milter_api/smfi_getsymval.html>`_
	from `libmilter` for more information.

	Warning:
		The current implementation is very naïve and does not behave exactly like
		`libmilter`, nor is it very robust.  It will definitely change in the future.
	"""

	headers: HeadersAccessor
	"""
	A `HeadersAccessor` object for accessing and modifying the message header fields
	"""

	body: BodyAccessor
	"""
	A `BodyAccessor` object for accessing and modifying the message body
	"""

	def __init__(
		self,
		sender: Sender,
		broadcast: util.Broadcast[EventMessage]|None = None,
	):
		self.host = ""
		self.address = None
		self.port = 0

		self.sender = sender
		self.broadcast = broadcast or util.Broadcast[EventMessage]()

		self.macros = dict[str, str]()
		self.headers = HeadersAccessor(self, sender)
		self.body = BodyAccessor(self, sender)

		# Phase checking is a bit fuzzy as a filter may not request every message,
		# so some phases will be skipped; checks should not try to exactly match a phase.
		self.phase = Phase.INIT

		self._helo: Helo|None = None

	async def __aenter__(self) -> Self:
		await self.broadcast.__aenter__()
		return self

	async def __aexit__(self, *_: object) -> None:
		await self.broadcast.__aexit__(None, None, None)
		# on session close, wake up any remaining deliver() awaitables
		await self.broadcast.shutdown_hook()

	def _reset(self) -> None:
		self.headers = HeadersAccessor(self, self.sender)
		self.body = BodyAccessor(self, self.sender)

	async def deliver(self, message: EventMessage) -> type[Continue]|type[Skip]:
		"""
		Deliver a message (or its contents) to a task waiting for it
		"""
		match message:
			case Connect():
				self.host = message.hostname
				self.address = message.address
				self.port = message.port
				async with self.broadcast:
					self.phase = Phase.CONNECT
				return Continue
			case Macro():
				self.macros.update(message.macros)
				return Continue  # not strictly necessary, but type checker needs something
			case Abort():
				async with self.broadcast:
					self.phase = Phase.CONNECT
				await self.broadcast.abort(Aborted)
				self._reset()
				return Continue
			case Helo():
				phase = Phase.MAIL
			case EnvelopeFrom() | EnvelopeRecipient() | Unknown():
				phase = Phase.ENVELOPE
			case Data() | Header():
				phase = Phase.HEADERS
			case EndOfHeaders() | Body():
				phase = Phase.BODY
			case EndOfMessage():  # pragma: no-branch
				phase = Phase.POST
		async with self.broadcast:
			self.phase = phase  # phase attribute must be modified in locked context
		await self.broadcast.send(message)
		return Skip if self.phase == Phase.BODY and self.body.should_skip() else Continue

	async def helo(self) -> str:
		"""
		Wait for a HELO/EHLO message and return the client's claimed hostname
		"""
		if self.phase > Phase.CONNECT:
			raise RuntimeError(
				"Session.helo() must be awaited before any other async features of a "
				"Session",
			)
		if self._helo:
			return self._helo.hostname
		while self.phase <= Phase.CONNECT:
			message = await self.broadcast.receive()
			if isinstance(message, Helo):
				self._helo = message
				return message.hostname
		raise RuntimeError("HELO/EHLO event not received")

	async def envelope_from(self) -> str:
		"""
		Wait for a MAIL command message and return the sender identity

		Note that if extensions arguments are wanted, users should use `Session.extension()`
		instead with a name of ``"MAIL"``.
		"""
		if self.phase > Phase.MAIL:
			raise RuntimeError(
				"Session.envelope_from() may only be awaited before the ENVELOPE phase",
			)
		while self.phase <= Phase.MAIL:
			message = await self.broadcast.receive()
			if isinstance(message, EnvelopeFrom):
				return bytes(message.sender).decode()
		raise RuntimeError("MAIL event not received")

	async def envelope_recipients(self) -> AsyncIterator[str]:
		"""
		Wait for RCPT command messages and iteratively yield the recipients' identities

		Note that if extensions arguments are wanted, users should use `Session.extension()`
		instead with a name of ``"RCPT"``.
		"""
		if self.phase > Phase.ENVELOPE:
			raise RuntimeError(
				"Session.envelope_from() may only be awaited before the HEADERS phase",
			)
		while self.phase <= Phase.ENVELOPE:
			message = await self.broadcast.receive()
			if isinstance(message, EnvelopeRecipient):
				yield bytes(message.recipient).decode()

	async def extension(self, name: str) -> memoryview:
		"""
		Wait for the named command extension and return the raw command for processing
		"""
		if self.phase > Phase.ENVELOPE:
			raise RuntimeError(
				"Session.extension() may only be awaited before the HEADERS phase",
			)
		bname = name.encode("utf-8")
		while self.phase <= Phase.ENVELOPE:
			message = await self.broadcast.receive()
			match message:
				case Unknown():
					if message.content[:len(bname)] == bname:
						assert isinstance(message.content, memoryview)
						return message.content
				# fake buffers for MAIL and RCPT commands
				case EnvelopeFrom() if name == "MAIL":
					vals = [b"MAIL FROM", message.sender, *message.arguments]
					return memoryview(b" ".join(vals))
				case EnvelopeRecipient() if name == "RCPT":
					vals = [b"RCPT TO", message.recipient, *message.arguments]
					return memoryview(b" ".join(vals))
		raise RuntimeError(f"{name} event not received")

	async def change_sender(self, sender: str, args: str = "") -> None:
		"""
		Move onto the `Phase.POST` phase and instruct the MTA to change the sender address
		"""
		await _until_editable(self)
		await self.sender.send(ChangeSender(sender, args or None))

	async def add_recipient(self, recipient: str, args: str = "") -> None:
		"""
		Move onto the `Phase.POST` phase and instruct the MTA to add a new recipient address
		"""
		await _until_editable(self)
		await self.sender.send(
			AddRecipientPar(recipient, args) if args else AddRecipient(recipient),
		)

	async def remove_recipient(self, recipient: str) -> None:
		"""
		Move onto the `Phase.POST` phase and instruct the MTA to remove a recipient address
		"""
		await _until_editable(self)
		await self.sender.send(RemoveRecipient(recipient))


class HeadersAccessor(AsyncContextManager["HeaderIterator"]):
	"""
	A class that allows access and modification of the message headers sent from an MTA

	To access headers (which are only available iteratively), use an instance as an
	asynchronous context manager; a `HeaderIterator` is returned when the context is
	entered.
	"""

	def __init__(self, session: Session, sender: Sender):
		self.session = session
		self.sender = sender
		self._table = list[Header]()
		self._aiter = ContextVar[HeaderIterator|None]("header-iter")

	async def __aenter__(self) -> HeaderIterator:
		if not (aiter := self._aiter.get(None)):
			aiter = HeaderIterator(self.__aiter())
			self._aiter.set(aiter)
		return aiter

	async def __aexit__(self, *_: object) -> None:
		if aiter := self._aiter.get():
			await aiter.aclose()
		self._aiter.set(None)

	async def __aiter(self) -> AsyncGenerator[Header, None]:
		# yield from cached headers first; allows multiple tasks to access the headers
		# in an uncoordinated manner; note the broadcaster is locked at this point
		for header in self._table:
			yield header
		seen = set(id(header) for header in self._table)
		while self.session.phase <= Phase.HEADERS:
			match (await self.session.broadcast.receive()):
				case Header() as header:
					header.freeze()
					self._table.append(header)
					seen.add(id(header))
					try:
						yield header
					except GeneratorExit:
						await self.collect()
						raise
				case EndOfHeaders():
					return
		# It's possible for collect() to have been called while yielded, in which case the
		# loop will end. Yield any headers that were stored by collect() but not yet
		# yielded.
		for header in self._table:
			if id(header) not in seen:
				yield header

	async def collect(self) -> None:
		"""
		Collect all headers without producing an iterator

		Calling this method before the `Phase.BODY` phase allows later processing of headers
		(after the HEADER phase) without the need for an empty loop.
		"""
		# note the similarities between this and __aiter; the difference is no mutex or
		# yields
		while self.session.phase <= Phase.HEADERS:
			match (await self.session.broadcast.receive()):
				case Header() as header:
					header.freeze()
					self._table.append(header)
				case _:
					return

	async def delete(self, header: Header) -> None:
		"""
		Move onto the `Phase.POST` phase and Instruct the MTA to delete the given header
		"""
		await self.collect()
		await _until_editable(self.session)
		index = _index_by_name(self._table, header)
		await self.sender.send(ChangeHeader(index, header.name, b""))
		self._table.remove(header)

	async def update(self, header: Header, value: bytes) -> None:
		"""
		Move onto the `Phase.POST` phase and Instruct the MTA to modify the value of a header
		"""
		await self.collect()
		await _until_editable(self.session)
		index = _index_by_name(self._table, header)
		await self.sender.send(ChangeHeader(index, header.name, value))
		index = self._table.index(header)
		self._table[index].value = value

	async def insert(self, header: Header, position: Position) -> None:
		"""
		Move onto the `Phase.POST` phase and instruct the MTA to insert a new header

		The header is inserted at `START`, `END`, or a relative position with `Before` and
		`After`; for example ``Before(Header("To", "test@example.com"))``.
		"""
		await self.collect()
		await _until_editable(self.session)
		match position:
			case Position(subject="start"):
				index = 0
			case Position(subject="end"):
				index = len(self._table)
			case Before():
				index = self._table.index(position.subject)
			case After():  # pragma: no-branch
				index = self._table.index(position.subject) + 1
			case _:
				raise TypeError("Expect a Position")
		if index >= len(self._table):
			await self.sender.send(AddHeader(header.name, header.value))
			self._table.append(header)
		else:
			await self.sender.send(InsertHeader(index + 1, header.name, header.value))
			self._table.insert(index, header)


class HeaderIterator(AsyncGenerator[Header, None]):
	"""
	Iterator for headers obtained by using a `HeadersAccessor` as a context manager
	"""

	def __init__(self, aiter: AsyncGenerator[Header, None]):
		self._aiter = aiter

	def __aiter__(self) -> Self:
		return self

	async def __anext__(self) -> Header:  # noqa: D102
		return await self._aiter.__anext__()

	async def asend(self, value: None = None) -> Header:  # noqa: D102
		return await self._aiter.__anext__()

	async def athrow(  # noqa: D102
		self,
		e: type[BaseException]|BaseException,
		m: object = None,
		t: TracebackType|None = None, /,
	) -> Header:
		if isinstance(e, type):
			return await self._aiter.athrow(e, m, t)
		assert m is None
		return await self._aiter.athrow(e, m, t)

	async def aclose(self) -> None:  # noqa: D102
		await self._aiter.aclose()

	async def restrict(self, *names: str) -> AsyncIterator[Header]:
		"""
		Return an asynchronous generator that filters headers by name
		"""
		async for header in self._aiter:
			if header.name in names:
				yield header


class BodyAccessor(AsyncContextManager[AsyncIterator[memoryview]]):
	"""
	A class that allows access and modification of the message body sent from an MTA

	To access chunks of a body (which are only available iteratively), use an instance as an
	asynchronous context manager; an asynchronous iterator is returned when the context is
	entered.
	"""

	def __init__(self, session: Session, sender: Sender):
		self.session = session
		self.sender = sender
		self._entered = 0
		self._skip = False
		self._aiter = ContextVar[AsyncGenerator[memoryview, None] | None]("body-iter")

	async def __aenter__(self) -> AsyncIterator[memoryview]:
		if not (aiter := self._aiter.get(None)):
			aiter = self.__aiter()
			self._aiter.set(aiter)
		self._entered += 1
		return aiter

	async def __aexit__(self, *_: object) -> None:
		if aiter := self._aiter.get(None):
			await aiter.aclose()
		self._aiter.set(None)
		self._entered -= 1

	async def __aiter(self) -> AsyncGenerator[memoryview, None]:
		while self.session.phase <= Phase.BODY:
			match (await self.session.broadcast.receive()):
				case Body() as body:
					assert isinstance(body.content, memoryview)
					yield body.content
				case EndOfMessage() as eom:
					assert isinstance(eom.content, memoryview)
					yield eom.content

	def should_skip(self) -> bool:
		"""
		Return whether the message body should be skipped

		The body should be skipped when there are no active contexts.  All correctly
		implemented filters should have started a context before the first `Body` message.

		Once this method returns `True` it becomes "locked in" and will always return `True`
		after.
		"""
		if self._skip:
			return True
		self._skip = self._entered == 0
		return self._skip

	async def write(self, chunk: bytes) -> None:
		"""
		Request that chunks of a new message body are sent to the MTA

		This method should not be called from within the scope created by using it's
		instance as an async context (`async with`); doing so may cause a warning to be
		issued and the rest of the message body to be skipped.
		"""
		if self._aiter.get(None):
			warn(
				"it looks as if BodyAccessor.write() was called on an instance from within "
				"it's own async context",
				stacklevel=2,
			)
		await _until_editable(self.session)
		await self.sender.send(ReplaceBody(chunk))


async def _until_editable(session: Session) -> None:
	if session.phase == Phase.POST:
		return
	while session.phase < Phase.POST:
		if session.phase == Phase.HEADERS:
			await session.headers.collect()
		else:
			await session.broadcast.receive()


def _index_by_name(table: Sequence[Header], needle: Header) -> int:
	index = 0
	name = needle.name.lower()
	for header in table:
		if header == needle:
			return index + 1
		if header.name.lower() == name:
			index += 1
	raise ValueError(f"header not found: {needle}")
